#!/usr/bin/env python
# -*- coding: utf-8 -*-

# geoarray, A fast Python interface for image geodata - either on disk or in memory.
#
# Copyright (C) 2017-2023
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
test_geoarray
-------------

Tests for the geoarray.baseclasses module.
"""
import zipfile
from unittest import TestCase
import tempfile
import os
from collections import OrderedDict
from typing import Iterable
from tempfile import TemporaryDirectory
from difflib import unified_diff
from subprocess import getoutput

import numpy as np
from shapely.geometry import Polygon
from osgeo.osr import SpatialReference  # noqa
import matplotlib
from parameterized import parameterized  # noqa
from py_tools_ds.geo.vector import geometry
from py_tools_ds.geo.coord_trafo import transform_any_prj

from geoarray import GeoArray, masks, __file__ as geoarray_root  # noqa
from geoarray.metadata import GDAL_Metadata


__author__ = ["Daniel Scheffler", "Jessica Palka"]


# Path of the tests-directory in the geoarray-package.
tests_path = os.path.abspath(os.path.join(geoarray_root, "..", ".."))
path_data_tif = os.path.join(tests_path, "tests", "data", "L8_2bands_extract10x11.tif")
path_subset_bsq = os.path.join(tests_path, "tests", "data", "subset_metadata.bsq")
path_subset_hdr = os.path.join(tests_path, "tests", "data", "subset_metadata.hdr")
path_subset_tif = os.path.join(tests_path, "tests", "data", "subset_metadata.tif")

# disable matplotlib figure popups
matplotlib.use('Template')  # disables matplotlib figure popups


def _get_gA_inMem_notInMem(path_file=path_data_tif):
    gA_inMem = GeoArray(path_file)
    gA_inMem.to_mem()
    gA_inMem.filePath = ""

    gA_notInMem = GeoArray(path_file)

    params_inMem_NotInMem = [("inMem", gA_inMem),
                             ("notInMem", gA_notInMem)]

    return params_inMem_NotInMem


class Test_GeoArray(TestCase):
    # expected results
    given_bandnames = ['B1', 'B2']

    @classmethod
    def setUpClass(cls) -> None:
        cls.gA = _get_gA_inMem_notInMem()[0][1]

    def test_vsi_path(self):
        os.environ['CPL_ZIP_ENCODING'] = 'UTF-8'  # avoids a GDAL warning

        with tempfile.TemporaryDirectory() as td:
            path_zip = os.path.join(td, 'temp_archive.zip')
            file_basename = os.path.basename(path_data_tif)

            with zipfile.ZipFile(path_zip, 'w') as zf:
                zf.write(path_data_tif, file_basename)

            gA = GeoArray(f'/vsizip/{path_zip}/{file_basename}')
            self.assertIsInstance(gA[:], np.ndarray)

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_bandnames(self, _, gA):
        """Test GeoArray.bandnames."""
        self.assertEqual(gA.bandnames, OrderedDict([('B1', 0), ('B2', 1)]))

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_shape(self, _, gA):
        """Test GeoArray.shape, .ndim, .rows, .columns, and .bands"""
        shape_attrs = (gA.ndim, gA.rows, gA.columns, gA.bands)
        shape_property = ('DIMENSIONS', 'ROWS', 'COLUMNS', 'BANDS')
        R_exp, C_exp, B_exp = 10, 11, 2
        expected_result = (3, R_exp, C_exp, B_exp)  # dimensions, rows, columns, band

        self.assertEqual(gA.shape, (R_exp, C_exp, B_exp))

        for i in range(4):
            with self.subTest(i=i):
                self.assertEqual(shape_attrs[i], expected_result[i],
                                 msg='The number of {i} is different from the expected result!'.format(
                                     i=shape_property[i]))

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_shape_is_updated(self, _, gA):
        """Test if shape attribute gets properly updated if the full array is replaced with one with another shape."""
        gA.arr = np.random.randint(1, 10, (gA.shape[0], ))
        self.assertEqual((gA.shape[0],), gA.shape, 'Shape attribute was not correctly updated.')
        self.assertEqual((gA.shape[0],), gA._shape, 'Private shape attribute was not correctly updated.')

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_dtype(self, _, gA):
        """Test GeoArray.dtype."""
        self.assertEqual(gA.dtype, np.dtype('float32'))

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_dtype_is_updated(self, _, gA):
        """Test if the data type gets properly updated if the full array is replaced with one with another data type."""
        self.assertEqual(gA.dtype, np.float32)

        gA.arr = np.random.randint(1, 10, (10, 10, 2), int)
        self.assertEqual(gA.dtype, int, 'Data type attribute was not correctly updated.')
        self.assertEqual(gA._dtype, int, 'Private data type attribute was not correctly updated.')

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_geotransform_gsd_grid(self, _, gA):
        """Test GeoArray.geotransform, .xgsd, .ygsd, and .xygrid_specs."""
        self.assertEqual(gA.geotransform, [365985.0, 30.0, 0.0, 5916615.0, 0.0, -30.0])
        self.assertEqual((gA.xgsd, gA.ygsd), (30, 30))
        self.assertEqual(gA.xygrid_specs, [[365985.0, 366015.0], [5916615.0, 5916585.0]],
                         msg='The [[xOrigin, xGSD], [yOrigin, yGSD]]-grid is not as expected!')

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_projection_epsg(self, _, gA):
        """Test GeoArray.projection and .epsg."""

        # Convert WKT-string of the projection to a Proj4_string
        srs = SpatialReference()
        srs.ImportFromWkt(gA.projection)
        proj4 = srs.ExportToProj4()

        self.assertEqual(proj4.strip(' /t/n/r'), '+proj=utm +zone=33 +datum=WGS84 +units=m +no_defs')
        self.assertEqual(gA.epsg, 32633)

    def test_nodata_not_given(self):
        """Test GeoArray.nodata in case no nodata value is given (i.e., auto-detected)."""
        gA = GeoArray(np.random.randint(1, 10, (5, 10)))
        self.assertIsNone(gA._nodata)
        self.assertEqual(self.gA.nodata, -9999.0)

    def test_nodata_given(self):
        """Test GeoArray.nodata in case nodata value is given."""
        gA = GeoArray(np.array([1, 2]), nodata=-9999)
        self.assertEqual(gA.nodata, -9999.0)

    def test_nodata_out_of_range(self):
        """Test if given nodata value is valid with respect to the array data type."""
        # test nodata value in range
        GeoArray(np.array([1, 2]).astype(np.uint8), nodata=255)

        # test nodata value out of range in range
        with self.assertRaises(ValueError):
            GeoArray(np.array([1, 2]).astype(np.uint8), nodata=256)

        # test float nodata value together with integer array data type
        GeoArray(np.array([1, 2]).astype(np.uint8), nodata=0.)

        # test NaN nodata value together with float array data type
        GeoArray(np.array([1, 2]).astype(float), nodata=np.nan)

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_calc_mask_nodata(self, _, gA):
        def test_with_flag_all(_gA):
            """Test mask output with flag='all' (flag when nodata in ALL bands)."""
            mnd_all = _gA.calc_mask_nodata(overwrite=True, flag='all')
            mnd_all_exp = np.ones((10, 11)).astype(bool)
            mnd_all_exp[0, 0] = False  # False represents nodata
            self.assertTrue(np.array_equal(mnd_all, mnd_all_exp))

        def test_with_flag_any(_gA):
            """Test mask output with flag='any' (flag when nodata in ANY band)."""
            mnd_any = _gA.calc_mask_nodata(overwrite=True, flag='any')
            mnd_any_exp = np.ones((10, 11)).astype(bool)
            mnd_any_exp[0, 0:3] = False  # False represents nodata
            self.assertTrue(np.array_equal(mnd_any, mnd_any_exp))

        if gA.is_inmem:
            _real_nodata = gA.nodata
            try:
                for nodata in [_real_nodata, np.nan]:
                    gA.nodata = nodata

                    gA[0, 0, :] = nodata  # set pixel [0, 0] to nodata in all bands
                    gA[0, 1, 0] = nodata  # set pixel [0, 1] to nodata in first band only
                    gA[0, 2, 1] = nodata  # set pixel [0, 2] to nodata in second band only

                    test_with_flag_all(gA)
                    test_with_flag_any(gA)

            finally:
                gA.nodata = _real_nodata

            with self.assertRaises(ValueError):
                gA.calc_mask_nodata(overwrite=True, flag='bad_string')

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test___getitem__(self, _, gA):
        def validate(array, exp_shape):
            self.assertIsInstance(array, np.ndarray)
            self.assertEqual(array.shape, exp_shape)

        R, C, B = gA.shape  # (10, 11, 2)

        # test row/col subset
        validate(gA[:1, :3, :], (1, 3, B))  # only one row is requested, given as a slice
        validate(gA[0, :3, :], (3, B))  # only one row is requested, given as an int
        validate(gA[2:5, :3], (3, 3, B))  # third dimension is not given
        validate(gA[2:5, :3, :], (3, 3, B))

        # test band subset
        validate(gA[:, :, 0:1], (R, C, 1))  # band slice  # returns 3D array
        validate(gA[:, :, 0], (R, C))  # band indexing  # returns 2D array
        validate(gA[1], (R, C))  # only band is given  # returns 2D
        validate(gA['B1'], (R, C))  # only bandname is given

        # test read point data in case only data of a single band is needed
        validate(gA[[1, 5], [2, 4], 1], (2, ))
        if not gA.is_inmem:
            self.assertNotEqual(gA._arr_cache['arr_cached'].shape, gA.shape)

        # test wrong inputs
        self.assertRaises(ValueError, gA.__getitem__, 'B01')

        # test full array  # NOTE: This sets self.gA.arr!
        validate(gA[:], (R, C, B))

        # TODO: add further tests

    def test___getitem__consistency(self):
        testarr = np.zeros((2, 2, 2))
        testarr[0, :, :] = [[11, 12], [13, 14]]
        testarr[1, :, :] = [[21, 22], [23, 24]]

        gA_inmem = GeoArray(testarr)
        inmem_res = gA_inmem[0, :, :]

        with tempfile.TemporaryDirectory() as tf:
            gA_inmem.save(os.path.join(tf, 'test.bsq'))

            gA_notinmem = GeoArray(os.path.join(tf, 'test.bsq'))
            notinmem_res = gA_notinmem[0, :, :]

        self.assertEqual(inmem_res.ndim, notinmem_res.ndim)
        self.assertEqual(inmem_res.shape, notinmem_res.shape)

    def test___getitem__consistency_3d_array_1_column(self):
        testarr = np.array([[1, 2], [3, 4]]).reshape((2, 1, 2))

        gA_inmem = GeoArray(testarr)
        inmem_res = gA_inmem[:]

        with tempfile.TemporaryDirectory() as tf:
            gA_inmem.save(os.path.join(tf, 'test.bsq'))

            gA_notinmem = GeoArray(os.path.join(tf, 'test.bsq'))
            notinmem_res = gA_notinmem[:]

        self.assertEqual(inmem_res.ndim, notinmem_res.ndim)
        self.assertEqual(inmem_res.shape, notinmem_res.shape)

    def test___getitem__consistency_2d_array(self):
        testarr = np.zeros((2, 2))
        testarr[:, :] = [[11, 12], [13, 14]]

        gA_inmem = GeoArray(testarr)
        inmem_res = gA_inmem[0, 0]

        with tempfile.TemporaryDirectory() as tf:
            gA_inmem.save(os.path.join(tf, 'test.bsq'))

            gA_notinmem = GeoArray(os.path.join(tf, 'test.bsq'))
            notinmem_res = gA_notinmem[0, 0]

        self.assertEqual(inmem_res.ndim, notinmem_res.ndim)
        self.assertEqual(inmem_res.shape, notinmem_res.shape)

    def test_numpy_array(self):
        arr = np.array(self.gA)
        self.assertIsInstance(arr, np.ndarray)
        self.assertEqual(arr.shape, self.gA.shape)

    def test_box(self):
        """Test GeoArray.box."""
        self.assertIsInstance(self.gA.box, geometry.boxObj)

    def test_mask_nodata(self):
        # TODO: Consider the dependency of mask_nodata on calc_mask_nodata().
        """Test GeoArray.mask_nodata."""
        self.assertIsInstance(self.gA.mask_nodata, masks.NoDataMask)

    def test_mask_baddata(self):
        """Test GeoArray.mask_baddata."""
        bdm = masks.BadDataMask(os.path.join(tests_path, "tests", "data", "L8_BadDataMask10x11.tif"))

        for i in range(2):
            with self.subTest(i=1):
                if i == 0:
                    # mask_baddata-function is "None", when it is not actively set
                    self.assertIsNone(self.gA.mask_baddata)
                if i == 1:
                    self.gA.mask_baddata = bdm
                    self.assertIsInstance(self.gA.mask_baddata, masks.BadDataMask)

    def test_footprint_poly(self):
        # TODO: Test the validation of the footprint_poly-function.
        # TODO: Consider the dependencies of the footprint_poly-function on mask_nodata, boxObj.
        """Test GeoArray.footprint_poly."""
        self.assertIsInstance(self.gA.footprint_poly, Polygon)

    def test_clip_to_footprint(self):
        # 3D GeoArray
        gA_3D = GeoArray(path_data_tif).to_mem()  # (10, 11, 2)
        gA_3D[:, 0, :] = -9999  # set first column to nodata so that the footprint is a bit smaller
        gA_3D.clip_to_footprint()
        self.assertEqual(gA_3D.shape, (10, 10, 2))

        # 2D GeoArray
        _gA_3D = GeoArray(path_data_tif).to_mem()
        gA_2D = GeoArray(_gA_3D[0], _gA_3D.gt, _gA_3D.prj, nodata=-9999)  # (10, 11)
        gA_2D[:, 0] = -9999  # set first column to nodata so that the footprint is a bit smaller
        gA_2D.clip_to_footprint()
        self.assertEqual(gA_2D.shape, (10, 10))

    def test_metadata(self):
        # TODO: Create a metadata-file for the tested TIFF-Image.
        # TODO: Test, if the metadata-function gives an output
        """Test GeoArray.metadata."""
        self.assertIsInstance(self.gA.metadata, GDAL_Metadata)

    def test_tiles(self):
        test_gAs = [self.gA,  # 3D
                    GeoArray(self.gA[:, :, 0], geotransform=self.gA.gt, projection=self.gA.prj)]  # 2D

        for gA in test_gAs:
            tiles = gA.tiles(tilesize=(50, 50))
            self.assertIsInstance(tiles, Iterable)

            for ((rS, rE), (cS, cE)), tile in tiles:
                self.assertTrue(np.array_equal(tile, gA[rS: rE + 1, cS: cE + 1]))

    def test_get_subset_3D_geoarray(self):
        # test without resetting band names
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2))
        self.assertIsInstance(sub_gA, GeoArray)
        self.assertTrue(list(sub_gA.bandnames), list(self.gA.bandnames)[1])

        # test with providing only xslice
        sub_gA = self.gA.get_subset(xslice=slice(2, 5))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with providing only yslice
        sub_gA = self.gA.get_subset(yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with zslice provided as list
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=[0, 1])
        self.assertIsInstance(sub_gA, GeoArray)

        # test without providing zslice
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test requesting only one column
        sub_gA = self.gA.get_subset(xslice=slice(0, 1), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with resetting band names
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2),
                                    reset_bandnames=True)
        self.assertTrue(list(sub_gA.bandnames), ['B1'])

        # test arrays are equal
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2))
        sub_arr = self.gA[:3, 2:5, 1:2]
        self.assertTrue(np.array_equal(sub_gA[:], sub_arr))

        # test deepcopied arrays (modification of sub_gA.arr must not affect self.gA.arr)
        sub_gA[:2, :2] = 99
        self.assertTrue(np.array_equal(sub_gA[:2, :2], np.full((2, 2, 1), 99, self.gA.dtype)))
        self.assertNotEqual(np.mean(sub_arr[:2, :2]), 99)
        self.assertNotEqual(np.std(sub_arr[:2, :2]), 0)

        # test metadata
        self.assertEqual(sub_gA.meta.bands, 1)
        self.assertEqual(len(list(sub_gA.meta.band_meta.values())[0]), 1)
        self.assertEqual(len(list(sub_gA.bandnames.keys())), 1)
        self.assertNotEqual(sub_gA.gt, self.gA.gt)
        self.assertEqual(sub_gA.prj, self.gA.prj)

        # test not to return GeoArray
        out = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2),
                                 return_GeoArray=False)

        self.assertIsInstance(out, tuple)
        self.assertTrue(len(out), 3)
        self.assertIsInstance(out[0], np.ndarray)
        self.assertIsInstance(out[1], tuple)
        self.assertIsInstance(out[2], str)

    def test_get_subset_2D_geoarray(self):
        gA_2D = GeoArray(self.gA[0])

        # test without resetting band names
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)
        self.assertTrue(list(sub_gA.bandnames), list(self.gA.bandnames)[1])

        # test with providing only xslice
        sub_gA = gA_2D.get_subset(xslice=slice(2, 5))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with providing only yslice
        sub_gA = gA_2D.get_subset(yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test without providing zslice
        sub_gA = gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test requesting only one column
        sub_gA = self.gA.get_subset(xslice=slice(0, 1), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with resetting band names
        sub_gA = gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), reset_bandnames=True)
        self.assertTrue(list(sub_gA.bandnames), ['B1'])

        # test arrays are equal
        sub_gA = gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3))
        sub_gA_2D = gA_2D[:3, 2:5]
        self.assertTrue(np.array_equal(sub_gA[:], sub_gA_2D))

        # test deepcopied arrays (modification of sub_gA.arr must not affect self.gA.arr)
        sub_gA[:2, :2] = 99
        self.assertTrue(np.array_equal(sub_gA[:2, :2], np.full((2, 2), 99, gA_2D.dtype)))
        self.assertNotEqual(np.mean(sub_gA_2D[:2, :2]), 99)
        self.assertNotEqual(np.std(sub_gA_2D[:2, :2]), 0)

        # test metadata
        self.assertEqual(sub_gA.meta.bands, 1)
        self.assertEqual(len(list(sub_gA.meta.band_meta.values())[0]), 1)
        self.assertEqual(len(list(sub_gA.bandnames.keys())), 1)
        self.assertNotEqual(sub_gA.gt, gA_2D.gt)
        self.assertEqual(sub_gA.prj, gA_2D.prj)

        # test not to return GeoArray
        out = gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), return_GeoArray=False)

        # test with provided zslice
        with self.assertRaises(ValueError):
            gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2))
        with self.assertRaises(ValueError):
            gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=[1, 2])

        self.assertIsInstance(out, tuple)
        self.assertTrue(len(out), 3)
        self.assertIsInstance(out[0], np.ndarray)
        self.assertIsInstance(out[1], tuple)
        self.assertIsInstance(out[2], str)

    @parameterized.expand(_get_gA_inMem_notInMem())
    def test_save(self, _, gA):
        """Test GeoArray.save."""
        # save GeoArray and validate output
        for fmt in ['ENVI', 'GTiff']:

            with tempfile.TemporaryDirectory(dir=os.path.join(tests_path, "tests", "data", "output"),
                                             prefix='test_save_') as td:

                outpath = os.path.join(td, "TestGeoArray_10x11_copy.tif")
                gA.save(outpath, fmt=fmt)
                self.assertTrue(os.path.exists(outpath))

                gA_copy = GeoArray(outpath)
                self.assertIsInstance(gA_copy, GeoArray)
                self.assertTrue(np.array_equal(gA[:], gA_copy[:]))
                self.assertEqual(gA._nodata, gA_copy._nodata)

            # TODO validate written metadata (inMem, notInMem, in case a notInMem dataset already has custom metadata)

    @parameterized.expand(_get_gA_inMem_notInMem(path_subset_bsq))
    def test_save_meta_ENVI(self, _, gA):
        with TemporaryDirectory(prefix='geoarray__') as td:
            p_out = os.path.join(td, 'outfile.bsq')

            gA.save(p_out, fmt='ENVI')

            # compare gdalinfo output
            gdalinfo_out = [i + '\n' for i in getoutput(f'gdalinfo -mdd ENVI {p_out}').split('\n')]
            gdalinfo_orig = [i + '\n' for i in getoutput(f'gdalinfo -mdd ENVI {path_subset_bsq}').split('\n')]

            diff = ''.join(list(unified_diff(gdalinfo_orig[3:], gdalinfo_out[3:])))
            self.assertFalse(diff,
                             f'\n\nThe gdalinfo output of the saved dataset contains some unexpected differences: \n\n'
                             f'{diff}')

            # compare header files
            with open(p_out[:-4] + '.hdr', 'r') as inF:
                new = inF.readlines()
            with open(path_subset_hdr, 'r') as inF:
                old = inF.readlines()

            diff = ''.join(list(unified_diff(old[3:], new[3:])))
            self.assertFalse(diff)
            # os.system(f'gdalinfo {p_out}')

    @parameterized.expand(_get_gA_inMem_notInMem(path_subset_tif))
    def test_save_meta_GTiff(self, _, gA):
        with TemporaryDirectory(prefix='geoarray__') as td:
            p_out = os.path.join(td, 'outfile.tif')

            gA.save(p_out, fmt='GTiff')

            # compare gdalinfo output
            gdalinfo_out = [i + '\n' for i in getoutput(f'gdalinfo {p_out}').split('\n')]
            gdalinfo_orig = [i + '\n' for i in getoutput(f'gdalinfo {path_subset_tif}').split('\n')]

            diff = ''.join(list(unified_diff(gdalinfo_orig[2:], gdalinfo_out[2:])))
            self.assertFalse(diff,
                             f'\n\nThe gdalinfo output of the saved dataset contains some unexpected differences: \n\n'
                             f'{diff}')
            # os.system(f'gdalinfo {p_out}')

    def test_show(self):
        # test 3D case
        self.gA.show()
        self.gA.show(interactive=True)  # only works if test is started with IPython.

        # test 2D case
        gA = GeoArray(self.gA[:, :, 0])
        gA.show()
        with self.assertWarnsRegex(UserWarning, 'Currently there is no interactive mode for single-band arrays.'):
            gA.show(interactive=True)

    def test_show_map(self):
        self.gA.show_map()

    def test_show_map_noepsg(self):
        wkt_noepsg = \
            """
            PROJCRS["BU MEaSUREs Lambert Azimuthal Equal Area - SA - V01",
                BASEGEOGCRS["WGS 84",
                    DATUM["World Geodetic System 1984",
                        ELLIPSOID["WGS 84",6378137,298.257223563,
                            LENGTHUNIT["metre",1]]],
                    PRIMEM["Greenwich",0,
                        ANGLEUNIT["degree",0.0174532925199433]],
                    ID["EPSG",4326]],
                CONVERSION["unnamed",
                    METHOD["Lambert Azimuthal Equal Area",
                        ID["EPSG",9820]],
                    PARAMETER["Latitude of natural origin",-15,
                        ANGLEUNIT["degree",0.0174532925199433],
                        ID["EPSG",8801]],
                    PARAMETER["Longitude of natural origin",-60,
                        ANGLEUNIT["degree",0.0174532925199433],
                        ID["EPSG",8802]],
                    PARAMETER["False easting",0,
                        LENGTHUNIT["metre",1],
                        ID["EPSG",8806]],
                    PARAMETER["False northing",0,
                        LENGTHUNIT["metre",1],
                        ID["EPSG",8807]]],
                CS[Cartesian,2],
                    AXIS["easting",east,
                        ORDER[1],
                        LENGTHUNIT["metre",1]],
                    AXIS["northing",north,
                        ORDER[2],
                        LENGTHUNIT["metre",1]]]
            """
        wkt_noepsg = ' '.join(wkt_noepsg.split())

        gA = GeoArray(self.gA[:], self.gA.gt, wkt_noepsg)
        gA.show_map()

    def test_show_histogram(self):
        self.gA.show_histogram()

    def test_show_xprofile(self):
        nd_old = self.gA._nodata
        self.gA._nodata = self.gA[0, 0, 0]
        self.gA.show_xprofile(0, 0)
        self.gA.show_xprofile(0, 0, show_nodata=False)
        self.gA._nodata = nd_old

    def test_show_yprofile(self):
        self.gA.show_yprofile(0, 0)
        self.gA.show_yprofile(0, 0, show_nodata=False)

    def test_show_zprofile(self):
        self.gA.show_zprofile(0, 0)
        self.gA.show_zprofile(0, 0, show_nodata=False)
        self.gA.show_zprofile(0, 0, ylim=(0, 5))

        with self.assertRaises(RuntimeError):
            GeoArray(np.random.randint(1, 10, [10, 10])).show_zprofile(0, 0)

        if 'wavelength' not in self.gA.meta.band_meta:
            self.gA.meta.band_meta['wavelength'] = range(self.gA.bands)
            self.gA.show_zprofile(0, 0, ylim=(0, 5))
            del self.gA.meta.band_meta['wavelength']

    def test_get_mapPos(self):
        # TODO: validate results (pixel dimensions, values, metadata subset, ...)
        xmin, xmax, ymin, ymax = self.gA.box.boundsMap
        xgsd = self.gA.xgsd
        ygsd = self.gA.ygsd
        mapBounds = (xmin + xgsd, ymin, xmax + xgsd, ymax - ygsd / 2)

        arr_sub, gt, prj = self.gA.get_mapPos(mapBounds=mapBounds, mapBounds_prj=self.gA.prj)

        self.assertIsInstance(arr_sub, np.ndarray)
        self.assertIsInstance(gt, (tuple, list))
        self.assertIsInstance(prj, str)

        # -- validate that the result of get_mapPos is larger if mapBounds_prj is unequal to GeoArray.prj --

        # get map bounds in UTM zone 32
        xmin, ymin, xmax, ymax = mapBounds
        ULxy, URxy, LRxy, LLxy = \
            [transform_any_prj(self.gA.prj, 32632, X, Y)
             for X, Y in [(xmin, ymax), (xmax, ymax), (xmax, ymin), (xmin, ymin)]]
        xvals, yvals = zip(ULxy, URxy, LRxy, LLxy)
        mapBounds_32 = min(xvals), min(yvals), max(xvals), max(yvals)

        arr_sub_32, gt_32, prj_32 = self.gA.get_mapPos(mapBounds=mapBounds_32, mapBounds_prj=32632)

        self.assertIsInstance(arr_sub, np.ndarray)
        self.assertIsInstance(gt, (tuple, list))
        self.assertIsInstance(prj, str)
        self.assertTrue(arr_sub_32.size > arr_sub.size)

        arr_sub_32, gt_32, prj_32 = self.gA.get_mapPos(mapBounds=mapBounds_32, mapBounds_prj=32632, band2get=0)
        self.assertTrue(GeoArray(arr_sub_32, gt_32, prj_32).bands == 1)
        self.assertTrue(len(arr_sub_32.shape) == 2)

    def test_read_pointData__singleCoord_oneband(self):
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[366015, 5916585]]),
                                        band=1)

        self.assertEqual(vals_z.dtype, self.gA.dtype)
        self.assertEqual(vals_z, 5766)

    def test_read_pointData__singleCoord_allbands(self):
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[366015, 5916585]]))

        self.assertIsInstance(vals_z, np.ndarray)
        self.assertEqual(vals_z.size, self.gA.bands)
        self.assertTrue(np.array_equal(vals_z, np.array([[5352, 5766]])))

    def test_read_pointData__singleCoord_allbands_otherPrj(self):
        X, Y = transform_any_prj(self.gA.epsg, 4326, 366015.1, 5916584.9)
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[X, Y]]),
                                        mapXY_points_prj=4326)

        self.assertIsInstance(vals_z, np.ndarray)
        self.assertEqual(vals_z.size, self.gA.bands)
        self.assertTrue(np.array_equal(vals_z, np.array([[5352, 5766]])))

    def test_read_pointData__multiCoords_allbands(self):
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[366015, 5916585],
                                                               [366165, 5916400]]))

        self.assertIsInstance(vals_z, np.ndarray)
        shape_exp = (2, 1, self.gA.bands)
        self.assertEqual(vals_z.shape, shape_exp)
        self.assertTrue(np.array_equal(vals_z, np.array([[5352, 5766],
                                                         [5080, 5436]]).reshape(shape_exp)))

    def test_read_pointData__multiCoords_allbands_otherPrj(self):
        X, Y = transform_any_prj(self.gA.epsg, 4326,
                                 np.array([366015.1, 366165]),
                                 np.array([5916584.9, 5916399.9]))
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[X, Y]]),
                                        mapXY_points_prj=4326)

        self.assertIsInstance(vals_z, np.ndarray)
        shape_exp = (2, 1, self.gA.bands)
        self.assertEqual(vals_z.shape, shape_exp)
        self.assertTrue(np.array_equal(vals_z, np.array([[5352, 5766],
                                                         [5080, 5436]]).reshape(shape_exp)),
                        'vals_z: %s' % repr(vals_z))

    def test_read_pointData__singleCoord_offside_oneband(self):
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[365900, 5916585]]),
                                        band=1,
                                        offside_val=-9999)

        self.assertIsInstance(vals_z, int)
        self.assertEqual(vals_z, -9999)

    def test_read_pointData__multiCoords_offside_oneband(self):
        xtoolarge = self.gA.box.boundsMap[1] + self.gA.xgsd
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[365900, 5916585],  # x too small
                                                               [365900, 5917000],  # x too small, y to large
                                                               [xtoolarge, 5916585],  # x too large
                                                               [366015, 5916585],
                                                               [366165, 5916400]]),
                                        offside_val=-9999,
                                        band=1)

        self.assertIsInstance(vals_z, np.ndarray)
        shape_exp = (5, 1)
        self.assertEqual(vals_z.shape, shape_exp)
        self.assertTrue(np.array_equal(vals_z, np.array([[-9999],
                                                         [-9999],
                                                         [-9999],
                                                         [5766],
                                                         [5436]])))

    def test_read_pointData__multiCoords_offside_allbands(self):
        xtoolarge = self.gA.box.boundsMap[1] + self.gA.xgsd
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[365900, 5916585],  # x too small
                                                               [365900, 5917000],  # x too small, y to large
                                                               [xtoolarge, 5916585],  # x too large
                                                               [366015, 5916585],
                                                               [366165, 5916400]]),
                                        offside_val=-9999)

        self.assertIsInstance(vals_z, np.ndarray)
        shape_exp = (5, 1, self.gA.bands)
        self.assertEqual(vals_z.shape, shape_exp)
        self.assertTrue(np.array_equal(vals_z, np.array([[-9999, -9999],
                                                         [-9999, -9999],
                                                         [-9999, -9999],
                                                         [5352, 5766],
                                                         [5080, 5436]]).reshape(shape_exp)))


if __name__ == '__main__':
    import pytest
    pytest.main()
